﻿using System;
using SDX.Compiler;
using Mono.Cecil;
using Mono.Cecil.Cil;
using System.Linq;
using System.Collections.Generic;

public class PrefabSpawnRange : IPatcherMod
{
    public bool Link(ModuleDefinition gameModule, ModuleDefinition modModule)
    {
        InsertCW(gameModule, modModule);
        InsertOW(gameModule, modModule);
        InsertCopyPrefabData(gameModule, modModule);
        InsertPrefabInfoWZ(gameModule, modModule);
        InsertPrefabSpawnRuleKZ(gameModule, modModule);
        return true;
    }
    public void InsertPrefabInfoWZ(ModuleDefinition gameModule, ModuleDefinition modModule)
    {
        TypeDefinition classDef = gameModule.Types.First(d => d.Name == "PrefabInfo" && d.Namespace == "RWG2.Rules");
        MethodDefinition methodDef = classDef.Methods.First(d => d.Name == "WZ");

        Instruction insertPos = null;
        Instruction jumpCheck = null;
        Instruction jumpToPos = null;
        foreach (Instruction inst in methodDef.Body.Instructions)
        {
            if (inst.OpCode == OpCodes.Call && inst.Operand.GetType() == typeof(MethodReference) && ((MethodReference)inst.Operand).FullName == "!0 System.Collections.Generic.List`1/Enumerator<RWG2.Rules.PrefabInfo>::get_Current()")
            {
                jumpCheck = inst;
                insertPos = inst.Next.Next.Next.Next;
            }
            if (jumpCheck != null && inst.Operand != null && inst.Operand.GetType() == typeof(Instruction) && ((Instruction)inst.Operand).Offset == jumpCheck.Previous.Offset)
            {
                jumpToPos = inst.Previous.Previous;

            }
        }
        ILProcessor proc1 = methodDef.Body.GetILProcessor();

        TypeDefinition prefabSpawnRangeMethod = modModule.Types.First(d => d.Name == "PrefabSpawnRangeMethod");
        MethodDefinition isSpawn = prefabSpawnRangeMethod.Methods.First(d => d.Name == "IsSpawn");
        MethodReference isSpawnRef = gameModule.Import(isSpawn);
        proc1.InsertBefore(insertPos, Instruction.Create(OpCodes.Ldloc_2));
        proc1.InsertBefore(insertPos, Instruction.Create(OpCodes.Ldarg_3));
        proc1.InsertBefore(insertPos, Instruction.Create(OpCodes.Call, isSpawnRef));
        proc1.InsertBefore(insertPos, Instruction.Create(OpCodes.Brfalse, jumpToPos));
    }
    public void InsertPrefabSpawnRuleKZ(ModuleDefinition gameModule, ModuleDefinition modModule)
    {
        TypeDefinition classDef = gameModule.Types.First(d => d.Name == "PrefabSpawnRule" && d.Namespace == "RWG2.Rules");
        MethodDefinition methodDef = classDef.Methods.First(d => d.Name == "KZ");

        Instruction insertPos = null;
        Instruction jumpCheck = null;
        Instruction jumpToPos = null;
        foreach (Instruction inst in methodDef.Body.Instructions)
        {
            if (inst.OpCode == OpCodes.Call && inst.Operand.GetType() == typeof(MethodReference) && ((MethodReference)inst.Operand).FullName == "!0 System.Collections.Generic.List`1/Enumerator<RWG2.Rules.PrefabInfo>::get_Current()")
            {
                jumpCheck = inst;
                insertPos = inst.Next.Next;
            }
            if (jumpCheck != null && inst.Operand != null && inst.Operand.GetType() == typeof(Instruction) && ((Instruction)inst.Operand).Offset == jumpCheck.Previous.Offset)
            {
                jumpToPos = inst.Previous.Previous.Previous.Previous;

            }
        }
        ILProcessor proc1 = methodDef.Body.GetILProcessor();

        TypeDefinition prefabSpawnRangeMethod = modModule.Types.First(d => d.Name == "PrefabSpawnRangeMethod");
        MethodDefinition isSpawn = prefabSpawnRangeMethod.Methods.First(d => d.Name == "IsSpawn");
        MethodReference isSpawnRef = gameModule.Import(isSpawn);
        proc1.InsertBefore(insertPos, Instruction.Create(OpCodes.Ldloc_2));
        proc1.InsertBefore(insertPos, Instruction.Create(OpCodes.Ldarg_3));
        proc1.InsertBefore(insertPos, Instruction.Create(OpCodes.Call, isSpawnRef));
        proc1.InsertBefore(insertPos, Instruction.Create(OpCodes.Brfalse, jumpToPos));
    }

    public void InsertCopyPrefabData(ModuleDefinition gameModule, ModuleDefinition modModule)
    {
        TypeDefinition classDef = gameModule.Types.First(d => d.Name == "Prefab");
        MethodDefinition methodDef = classDef.Methods.First(d => d.Name == ".ctor" && d.Parameters.Count == 1);
        ILProcessor proc1 = methodDef.Body.GetILProcessor();
        Instruction insertPos = methodDef.Body.Instructions[0];

        TypeDefinition prefabSpawnRangeMethod = modModule.Types.First(d => d.Name == "PrefabSpawnRangeMethod");
        MethodDefinition copyPrefabData = prefabSpawnRangeMethod.Methods.First(d => d.Name == "CopyPrefabData");
        MethodReference copyPrefabDataRef = gameModule.Import(copyPrefabData);
        proc1.InsertBefore(insertPos, Instruction.Create(OpCodes.Ldarg_1));
        proc1.InsertBefore(insertPos, Instruction.Create(OpCodes.Ldarg_0));
        proc1.InsertBefore(insertPos, Instruction.Create(OpCodes.Call, copyPrefabDataRef));

    }


    public void InsertCW(ModuleDefinition gameModule, ModuleDefinition modModule)
    {
        TypeDefinition classDef = gameModule.Types.First(d => d.Name == "Prefab");
        MethodDefinition methodDef = classDef.Methods.First(d => d.Name == "CW");

        TypeDefinition prefabSpawnRangeMethod = modModule.Types.First(d => d.Name == "PrefabSpawnRangeMethod");
        MethodDefinition getPrefabSpawnRangeProperty = prefabSpawnRangeMethod.Methods.First(d => d.Name == "GetPrefabSpawnRangeProperty");
        MethodReference getPrefabSpawnRangePropertyRef = gameModule.Import(getPrefabSpawnRangeProperty);


        ILProcessor proc1 = methodDef.Body.GetILProcessor();

        Instruction insertPos = methodDef.Body.Instructions[0];

        proc1.InsertBefore(insertPos, Instruction.Create(OpCodes.Ldarg_0));
        proc1.InsertBefore(insertPos, Instruction.Create(OpCodes.Call, getPrefabSpawnRangePropertyRef));

    }

    public void InsertOW(ModuleDefinition gameModule, ModuleDefinition modModule)
    {
        TypeDefinition classDef = gameModule.Types.First(d => d.Name == "Prefab");
        MethodDefinition methodDef = classDef.Methods.First(d => d.Name == "OW");

        TypeDefinition prefabSpawnRangeMethod = modModule.Types.First(d => d.Name == "PrefabSpawnRangeMethod");
        MethodDefinition getPrefabSpawnRangeToProperty = prefabSpawnRangeMethod.Methods.First(d => d.Name == "GetPrefabSpawnRangeToProperty");
        MethodReference getPrefabSpawnRangeToPropertyRef = gameModule.Import(getPrefabSpawnRangeToProperty);


        ILProcessor proc1 = methodDef.Body.GetILProcessor();

        Instruction insertPos = methodDef.Body.Instructions[0];

        proc1.InsertBefore(insertPos, Instruction.Create(OpCodes.Ldarg_0));
        proc1.InsertBefore(insertPos, Instruction.Create(OpCodes.Call, getPrefabSpawnRangeToPropertyRef));

    }


    public void AddPrefabField(ModuleDefinition module)
    {
        TypeReference tr = module.Import(typeof(float));
        TypeDefinition prefab = module.Types.First(d => d.Name == "Prefab");

        tr = module.Import(typeof(float));
        AddFieldDefinition(prefab, "WorldSpawnRangeMin", FieldAttributes.Public, tr);
        AddFieldDefinition(prefab, "WorldSpawnRangeMax", FieldAttributes.Public, tr);
    }
    public bool Patch(ModuleDefinition module)
    {
        Console.WriteLine("==PrefabSpawnMax Patcher===");
        AddPrefabField(module);
        return true;
    }

    public void AddFieldDefinition(TypeDefinition classDef, string fieldName, FieldAttributes fieldAttribute, TypeReference fieldType)
    {
        FieldDefinition fd = new FieldDefinition(fieldName, fieldAttribute, fieldType);
        classDef.Fields.Add(fd);

    }
    public void AddFieldDefinition(TypeDefinition classDef, string fieldName, FieldAttributes fieldAttribute, TypeReference fieldType, object constantvalue)
    {
        FieldDefinition fd = new FieldDefinition(fieldName, fieldAttribute, fieldType);
        fd.Constant = constantvalue;
        classDef.Fields.Add(fd);

    }
    private void SetFieldToPublic(FieldDefinition field)
    {
        field.IsFamily = false;
        field.IsPrivate = false;
        field.IsPublic = true;

    }

}
